@@ -6,6 +6,7 @@
 
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 #include "memcached.h"
+#include "storage.h"
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/resource.h>
@@ -14,20 +15,22 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <signal.h>
 #include <string.h>
 #include <time.h>
 #include <assert.h>
 #include <unistd.h>
 #include <poll.h>
 
+#include "base64.h"
+
 #define LARGEST_ID POWER_LARGEST
 
 typedef struct {
     void *c; /* original connection structure. still with source thread attached. */
     int sfd; /* client fd. */
-    bipbuf_t *buf; /* output buffer */
-    char *cbuf; /* current buffer */
+    int buflen;
+    int bufused;
+    char *buf; /* output buffer */
 } crawler_client_t;
 
 typedef struct _crawler_module_t crawler_module_t;
@@ -51,6 +54,7 @@ struct _crawler_module_t {
     void *data; /* opaque data pointer */
     crawler_client_t c;
     crawler_module_reg_t *mod;
+    int status; /* flags/code/etc for internal module usage */
 };
 
 static int crawler_expired_init(crawler_module_t *cm, void *data);
@@ -64,28 +68,43 @@ crawler_module_reg_t crawler_expired_mod = {
     .doneclass = crawler_expired_doneclass,
     .finalize = crawler_expired_finalize,
     .needs_lock = true,
-    .needs_client = false
+    .needs_client = false,
 };
 
+static int crawler_metadump_init(crawler_module_t *cm, void *data);
 static void crawler_metadump_eval(crawler_module_t *cm, item *search, uint32_t hv, int i);
 static void crawler_metadump_finalize(crawler_module_t *cm);
 
 crawler_module_reg_t crawler_metadump_mod = {
-    .init = NULL,
+    .init = crawler_metadump_init,
     .eval = crawler_metadump_eval,
     .doneclass = NULL,
     .finalize = crawler_metadump_finalize,
     .needs_lock = false,
-    .needs_client = true
+    .needs_client = true,
 };
 
-crawler_module_reg_t *crawler_mod_regs[3] = {
-    &crawler_expired_mod,
-    &crawler_expired_mod,
-    &crawler_metadump_mod
+static int crawler_mgdump_init(crawler_module_t *cm, void *data);
+static void crawler_mgdump_eval(crawler_module_t *cm, item *search, uint32_t hv, int i);
+static void crawler_mgdump_finalize(crawler_module_t *cm);
+
+crawler_module_reg_t crawler_mgdump_mod = {
+    .init = crawler_mgdump_init,
+    .eval = crawler_mgdump_eval,
+    .doneclass = NULL,
+    .finalize = crawler_mgdump_finalize,
+    .needs_lock = false,
+    .needs_client = true,
 };
 
-static int lru_crawler_client_getbuf(crawler_client_t *c);
+crawler_module_reg_t *crawler_mod_regs[4] = {
+    &crawler_expired_mod,
+    &crawler_expired_mod,
+    &crawler_metadump_mod,
+    &crawler_mgdump_mod,
+};
+
+static int lru_crawler_write(crawler_client_t *c);
 crawler_module_t active_crawler_mod;
 enum crawler_run_type active_crawler_type;
 
@@ -106,14 +125,13 @@ static void *storage;
 
 /*** LRU CRAWLER THREAD ***/
 
-#define LRU_CRAWLER_WRITEBUF 8192
+#define LRU_CRAWLER_MINBUFSPACE 8192
 
 static void lru_crawler_close_client(crawler_client_t *c) {
     //fprintf(stderr, "CRAWLER: Closing client\n");
     sidethread_conn_close(c->c);
     c->c = NULL;
-    c->cbuf = NULL;
-    bipbuf_free(c->buf);
+    free(c->buf);
     c->buf = NULL;
 }
 
@@ -121,11 +139,20 @@ static void lru_crawler_release_client(crawler_client_t *c) {
     //fprintf(stderr, "CRAWLER: Closing client\n");
     redispatch_conn(c->c);
     c->c = NULL;
-    c->cbuf = NULL;
-    bipbuf_free(c->buf);
+    free(c->buf);
     c->buf = NULL;
 }
 
+static int lru_crawler_expand_buf(crawler_client_t *c) {
+    c->buflen *= 2;
+    char *nb = realloc(c->buf, c->buflen);
+    if (nb == NULL) {
+        return -1;
+    }
+    c->buf = nb;
+    return 0;
+}
+
 static int crawler_expired_init(crawler_module_t *cm, void *data) {
     struct crawler_expired_data *d;
     if (data != NULL) {
@@ -186,9 +213,7 @@ static void crawler_expired_eval(crawler_module_t *cm, item *search, uint32_t hv
 #ifdef EXTSTORE
     bool is_valid = true;
     if (search->it_flags & ITEM_HDR) {
-        item_hdr *hdr = (item_hdr *)ITEM_data(search);
-        if (extstore_check(storage, hdr->page_id, hdr->page_version) != 0)
-            is_valid = false;
+        is_valid = storage_validate_item(storage, search);
     }
 #endif
     if ((search->exptime != 0 && search->exptime < current_time)
@@ -236,9 +261,139 @@ static void crawler_expired_eval(crawler_module_t *cm, item *search, uint32_t hv
     pthread_mutex_unlock(&d->lock);
 }
 
+static int crawler_metadump_init(crawler_module_t *cm, void *data) {
+    cm->status = 0;
+    return 0;
+}
+
+#define KADD(p, s) { \
+    memcpy(p, s, sizeof(s)-1); \
+    p += sizeof(s)-1; \
+}
+#define KSP(p) { \
+    *p = ' '; \
+    p++; \
+}
+
 static void crawler_metadump_eval(crawler_module_t *cm, item *it, uint32_t hv, int i) {
-    //int slab_id = CLEAR_LRU(i);
-    char keybuf[KEY_MAX_URI_ENCODED_LENGTH];
+    int is_flushed = item_is_flushed(it);
+#ifdef EXTSTORE
+    bool is_valid = true;
+    if (it->it_flags & ITEM_HDR) {
+        is_valid = storage_validate_item(storage, it);
+    }
+#endif
+    /* Ignore expired content. */
+    if ((it->exptime != 0 && it->exptime < current_time)
+        || is_flushed
+#ifdef EXTSTORE
+        || !is_valid
+#endif
+        ) {
+        refcount_decr(it);
+        return;
+    }
+    client_flags_t flags;
+    FLAGS_CONV(it, flags);
+    assert(it->nkey * 3 < LRU_CRAWLER_MINBUFSPACE/2);
+    // unrolled snprintf for ~30% time improvement on fullspeed dump
+    // key=%s exp=%ld la=%llu cas=%llu fetch=%s cls=%u size=%lu flags=%llu\n
+    // + optional ext_page=%u and ext_offset=%u
+
+    char *p = cm->c.buf + cm->c.bufused;
+    char *start = p;
+    KADD(p, "key=");
+    p = uriencode_p(ITEM_key(it), p, it->nkey);
+    KSP(p);
+
+    KADD(p, "exp=");
+    if (it->exptime == 0) {
+        KADD(p, "-1 ");
+    } else {
+        p = itoa_64((long)(it->exptime + process_started), p);
+        KSP(p);
+    }
+
+    KADD(p, "la=");
+    p = itoa_u64((unsigned long long)(it->time + process_started), p);
+    KSP(p);
+
+    KADD(p, "cas=");
+    p = itoa_u64(ITEM_get_cas(it), p);
+    KSP(p);
+
+    if (it->it_flags & ITEM_FETCHED) {
+        KADD(p, "fetch=yes ");
+    } else {
+        KADD(p, "fetch=no ");
+    }
+
+    KADD(p, "cls=");
+    p = itoa_u32(ITEM_clsid(it), p);
+    KSP(p);
+
+    KADD(p, "size=");
+    p = itoa_u64(ITEM_ntotal(it), p);
+    KSP(p);
+
+    KADD(p, "flags=");
+    p = itoa_u64(flags, p);
+    KSP(p);
+
+#ifdef EXTSTORE
+    if (it->it_flags & ITEM_HDR) {
+#ifdef NEED_ALIGN
+    item_hdr hdr_s;
+    memcpy(&hdr_s, ITEM_data(it), sizeof(hdr_s));
+    item_hdr *hdr = &hdr_s;
+#else
+    item_hdr *hdr = (item_hdr *)ITEM_data(it);
+#endif
+
+    KADD(p, "ext_page=");
+    p = itoa_u32(hdr->page_id, p);
+    KSP(p);
+
+    KADD(p, "ext_offset=");
+    p = itoa_u32(hdr->offset, p);
+    KSP(p);
+    }
+#endif
+
+    KADD(p, "\n");
+
+    refcount_decr(it);
+    assert(p - start < LRU_CRAWLER_MINBUFSPACE-1);
+    cm->c.bufused += p - start;
+}
+
+#undef KADD
+#undef KSP
+
+static void crawler_metadump_finalize(crawler_module_t *cm) {
+    if (cm->c.c != NULL) {
+        // flush any pending data.
+        if (lru_crawler_write(&cm->c) == 0) {
+            // Only nonzero status right now means we were locked
+            if (cm->status != 0) {
+                const char *errstr = "ERROR locked try again later\r\n";
+                size_t errlen = strlen(errstr);
+                memcpy(cm->c.buf, errstr, errlen);
+                cm->c.bufused += errlen;
+            } else {
+                memcpy(cm->c.buf, "END\r\n", 5);
+                cm->c.bufused += 5;
+            }
+        }
+    }
+}
+
+static int crawler_mgdump_init(crawler_module_t *cm, void *data) {
+    cm->status = 0;
+    return 0;
+}
+
+static void crawler_mgdump_eval(crawler_module_t *cm, item *it, uint32_t hv, int i) {
     int is_flushed = item_is_flushed(it);
     /* Ignore expired content. */
     if ((it->exptime != 0 && it->exptime < current_time)
@@ -246,65 +401,84 @@ static void crawler_metadump_eval(crawler_module_t *cm, item *it, uint32_t hv, i
         refcount_decr(it);
         return;
     }
-    // TODO: uriencode directly into the buffer.
-    uriencode(ITEM_key(it), keybuf, it->nkey, KEY_MAX_URI_ENCODED_LENGTH);
-    int total = snprintf(cm->c.cbuf, 4096,
-            "key=%s exp=%ld la=%llu cas=%llu fetch=%s cls=%u size=%lu\n",
-            keybuf,
-            (it->exptime == 0) ? -1 : (long)(it->exptime + process_started),
-            (unsigned long long)(it->time + process_started),
-            (unsigned long long)ITEM_get_cas(it),
-            (it->it_flags & ITEM_FETCHED) ? "yes" : "no",
-            ITEM_clsid(it),
-            (unsigned long) ITEM_ntotal(it));
+
+    char *p = cm->c.buf + cm->c.bufused; // buffer offset.
+    char *start = p;
+    memcpy(p, "mg ", 3);
+    p += 3;
+    if (it->it_flags & ITEM_KEY_BINARY) {
+        p += base64_encode((unsigned char *) ITEM_key(it), it->nkey, (unsigned char*) p, LRU_CRAWLER_MINBUFSPACE/2);
+        memcpy(p, " b\r\n", 4);
+        p += 4;
+    } else {
+        memcpy(p, ITEM_key(it), it->nkey);
+        p += it->nkey;
+        memcpy(p, "\r\n", 2);
+        p += 2;
+    }
+    int total = p - start;
+
     refcount_decr(it);
-    // TODO: some way of tracking the errors. these are very unlikely though.
-    if (total >= LRU_CRAWLER_WRITEBUF - 1 || total <= 0) {
-        /* Failed to write, don't push it. */
-        return;
-    }
-    bipbuf_push(cm->c.buf, total);
+    cm->c.bufused += total;
 }
 
-static void crawler_metadump_finalize(crawler_module_t *cm) {
+static void crawler_mgdump_finalize(crawler_module_t *cm) {
     if (cm->c.c != NULL) {
-        // Ensure space for final message.
-        lru_crawler_client_getbuf(&cm->c);
-        memcpy(cm->c.cbuf, "END\r\n", 5);
-        bipbuf_push(cm->c.buf, 5);
+        // flush any pending data.
+        if (lru_crawler_write(&cm->c) == 0) {
+            // Only nonzero status right now means we were locked
+            if (cm->status != 0) {
+                const char *errstr = "ERROR locked try again later\r\n";
+                size_t errlen = strlen(errstr);
+                memcpy(cm->c.buf, errstr, errlen);
+                cm->c.bufused += errlen;
+            } else {
+                memcpy(cm->c.buf, "EN\r\n", 4);
+                cm->c.bufused += 4;
+            }
+        }
     }
 }
 
-static int lru_crawler_poll(crawler_client_t *c) {
-    unsigned char *data;
-    unsigned int data_size = 0;
+// write the whole buffer out to the client socket.
+static int lru_crawler_write(crawler_client_t *c) {
+    unsigned int data_size = c->bufused;
+    unsigned int sent = 0;
     struct pollfd to_poll[1];
     to_poll[0].fd = c->sfd;
     to_poll[0].events = POLLOUT;
 
-    int ret = poll(to_poll, 1, 1000);
+    if (c->c == NULL) return -1;
+    if (data_size == 0) return 0;
 
-    if (ret < 0) {
-        // fatal.
-        return -1;
-    }
+    while (sent < data_size) {
+        int ret = poll(to_poll, 1, 1000);
 
-    if (ret == 0) return 0;
-
-    if (to_poll[0].revents & POLLIN) {
-        char buf[1];
-        int res = ((conn*)c->c)->read(c->c, buf, 1);
-        if (res == 0 || (res == -1 && (errno != EAGAIN && errno != EWOULDBLOCK))) {
+        if (ret < 0) {
+            // fatal.
             lru_crawler_close_client(c);
             return -1;
         }
-    }
-    if ((data = bipbuf_peek_all(c->buf, &data_size)) != NULL) {
+
+        if (ret == 0) return 0;
+
+        // check if socket was closed on us.
+        if (to_poll[0].revents & POLLIN) {
+            char buf[1];
+            int res = ((conn*)c->c)->read(c->c, buf, 1);
+            if (res == 0 || (res == -1 && (errno != EAGAIN && errno != EWOULDBLOCK))) {
+                lru_crawler_close_client(c);
+                return -1;
+            }
+        }
+
         if (to_poll[0].revents & (POLLHUP|POLLERR)) {
+            // got socket hangup.
             lru_crawler_close_client(c);
             return -1;
         } else if (to_poll[0].revents & POLLOUT) {
-            int total = ((conn*)c->c)->write(c->c, data, data_size);
+            // socket is writeable.
+            int total = ((conn*)c->c)->write(c->c, c->buf + sent, data_size - sent);
             if (total == -1) {
                 if (errno != EAGAIN && errno != EWOULDBLOCK) {
                     lru_crawler_close_client(c);
@@ -313,29 +487,14 @@ static int lru_crawler_poll(crawler_client_t *c) {
             } else if (total == 0) {
                 lru_crawler_close_client(c);
                 return -1;
-            } else {
-                bipbuf_poll(c->buf, total);
             }
+            sent += total;
         }
-    }
-    return 0;
-}
+    } // while
 
-/* Grab some space to work with, if none exists, run the poll() loop and wait
- * for it to clear up or close.
- * Return NULL if closed.
- */
-static int lru_crawler_client_getbuf(crawler_client_t *c) {
-    void *buf = NULL;
-    if (c->c == NULL) return -1;
-    /* not enough space. */
-    while ((buf = bipbuf_request(c->buf, LRU_CRAWLER_WRITEBUF)) == NULL) {
-        // TODO: max loops before closing.
-        int ret = lru_crawler_poll(c);
-        if (ret < 0) return ret;
-    }
+    // write buffer now empty.
+    c->bufused = 0;
 
-    c->cbuf = buf;
     return 0;
 }
 
@@ -350,6 +509,87 @@ static void lru_crawler_class_done(int i) {
         active_crawler_mod.mod->doneclass(&active_crawler_mod, i);
 }
 
+// ensure we build the buffer a little bit to cut down on poll/write syscalls.
+#define MIN_ITEMS_PER_WRITE 16
+static void item_crawl_hash(void) {
+    // get iterator from assoc. can hang for a long time.
+    // - blocks hash expansion
+    void *iter = assoc_get_iterator();
+    int crawls_persleep = settings.crawls_persleep;
+    item *it = NULL;
+    int items = 0;
+
+    // Could not get the iterator: probably locked due to hash expansion.
+    if (iter == NULL) {
+        active_crawler_mod.status = 1;
+        return;
+    }
+
+    // loop while iterator returns something
+    // - iterator func handles bucket-walking
+    // - iterator returns with bucket locked.
+    while (assoc_iterate(iter, &it)) {
+        // if iterator returns true but no item, we're inbetween buckets and
+        // can do cleanup work without holding an item lock.
+        if (it == NULL) {
+            if (active_crawler_mod.c.c != NULL) {
+                if (items > MIN_ITEMS_PER_WRITE) {
+                    int ret = lru_crawler_write(&active_crawler_mod.c);
+                    items = 0;
+                    if (ret != 0) {
+                        // fail out and finalize.
+                        break;
+                    }
+                }
+            } else if (active_crawler_mod.mod->needs_client) {
+                // fail out and finalize.
+                break;
+            }
+
+            // - sleep bits from orig loop
+            if (crawls_persleep <= 0 && settings.lru_crawler_sleep) {
+                pthread_mutex_unlock(&lru_crawler_lock);
+                usleep(settings.lru_crawler_sleep);
+                pthread_mutex_lock(&lru_crawler_lock);
+                crawls_persleep = settings.crawls_persleep;
+            } else if (!settings.lru_crawler_sleep) {
+                // TODO: only cycle lock every N?
+                pthread_mutex_unlock(&lru_crawler_lock);
+                pthread_mutex_lock(&lru_crawler_lock);
+            }
+            continue;
+        }
+
+        // double check that the item isn't in a transitional state.
+        if (refcount_incr(it) < 2) {
+            refcount_decr(it);
+            continue;
+        }
+
+        // We're presently holding an item lock, so we cannot flush the
+        // buffer to the network socket as the syscall is both slow and could
+        // hang waiting for POLLOUT. Instead we must expand the buffer.
+        if (active_crawler_mod.c.c != NULL) {
+            crawler_client_t *c = &active_crawler_mod.c;
+            if (c->buflen - c->bufused < LRU_CRAWLER_MINBUFSPACE) {
+                if (lru_crawler_expand_buf(c) != 0) {
+                    // failed to expand buffer, stop.
+                    break;
+                }
+            }
+        }
+        // FIXME: missing hv and i are fine for metadump eval, but not fine
+        // for expire eval.
+        active_crawler_mod.mod->eval(&active_crawler_mod, it, 0, 0);
+        crawls_persleep--;
+        items++;
+    }
+
+    // must finalize or we leave the hash table expansion blocked.
+    assoc_iterate_final(iter);
+    return;
+}
+
 static void *item_crawler_thread(void *arg) {
     int i;
     int crawls_persleep = settings.crawls_persleep;
@@ -362,6 +602,10 @@ static void *item_crawler_thread(void *arg) {
     while (do_run_lru_crawler_thread) {
     pthread_cond_wait(&lru_crawler_cond, &lru_crawler_lock);
 
+    if (crawler_count == -1) {
+        item_crawl_hash();
+        crawler_count = 0;
+    } else {
     while (crawler_count) {
         item *search = NULL;
         void *hold_lock = NULL;
@@ -371,12 +615,14 @@ static void *item_crawler_thread(void *arg) {
                 continue;
             }
 
-            /* Get memory from bipbuf, if client has no space, flush. */
             if (active_crawler_mod.c.c != NULL) {
-                int ret = lru_crawler_client_getbuf(&active_crawler_mod.c);
-                if (ret != 0) {
-                    lru_crawler_class_done(i);
-                    continue;
+                crawler_client_t *c = &active_crawler_mod.c;
+                if (c->buflen - c->bufused < LRU_CRAWLER_MINBUFSPACE) {
+                    int ret = lru_crawler_write(c);
+                    if (ret != 0) {
+                        lru_crawler_class_done(i);
+                        continue;
+                    }
                 }
             } else if (active_crawler_mod.mod->needs_client) {
                 lru_crawler_class_done(i);
@@ -435,13 +681,14 @@ static void *item_crawler_thread(void *arg) {
                 pthread_mutex_lock(&lru_crawler_lock);
             }
         }
-    }
+    } // while
+    } // if crawler_count
 
     if (active_crawler_mod.mod != NULL) {
         if (active_crawler_mod.mod->finalize != NULL)
             active_crawler_mod.mod->finalize(&active_crawler_mod);
-        while (active_crawler_mod.c.c != NULL && bipbuf_used(active_crawler_mod.c.buf)) {
-            lru_crawler_poll(&active_crawler_mod.c);
+        while (active_crawler_mod.c.c != NULL && active_crawler_mod.c.bufused != 0) {
+            lru_crawler_write(&active_crawler_mod.c);
         }
         // Double checking in case the client closed during the poll
         if (active_crawler_mod.c.c != NULL) {
@@ -509,6 +756,7 @@ int start_item_crawler_thread(void) {
         pthread_mutex_unlock(&lru_crawler_lock);
         return -1;
     }
+    thread_setname(item_crawler_tid, "mc-itemcrawler");
     /* Avoid returning until the crawler has actually started */
     pthread_cond_wait(&lru_crawler_cond, &lru_crawler_lock);
     pthread_mutex_unlock(&lru_crawler_lock);
@@ -554,12 +802,6 @@ static int do_lru_crawler_start(uint32_t id, uint32_t remaining) {
         starts++;
     }
     pthread_mutex_unlock(&lru_locks[sid]);
-    if (starts) {
-        STATS_LOCK();
-        stats_state.lru_crawler_running = true;
-        stats.lru_crawler_starts++;
-        STATS_UNLOCK();
-    }
     return starts;
 }
 
@@ -571,10 +813,14 @@ static int lru_crawler_set_client(crawler_module_t *cm, void *c, const int sfd)
     crawlc->c = c;
     crawlc->sfd = sfd;
 
-    crawlc->buf = bipbuf_new(1024 * 128);
+    size_t size = LRU_CRAWLER_MINBUFSPACE * 16;
+    crawlc->buf = malloc(size);
+
     if (crawlc->buf == NULL) {
         return -2;
     }
+    crawlc->buflen = size;
+    crawlc->bufused = 0;
     return 0;
 }
 
@@ -605,6 +851,12 @@ int lru_crawler_start(uint8_t *ids, uint32_t remaining,
         return -1;
     }
 
+    /* hash table walk only supported with metadump for now. */
+    if (ids == NULL && type != CRAWLER_METADUMP && type != CRAWLER_MGDUMP) {
+        pthread_mutex_unlock(&lru_crawler_lock);
+        return -2;
+    }
+
     /* Configure the module */
     if (!is_running) {
         assert(crawler_mod_regs[type] != NULL);
@@ -625,12 +877,25 @@ int lru_crawler_start(uint8_t *ids, uint32_t remaining,
         }
     }
 
-    /* we allow the autocrawler to restart sub-LRU's before completion */
-    for (int sid = POWER_SMALLEST; sid < POWER_LARGEST; sid++) {
-        if (ids[sid])
-            starts += do_lru_crawler_start(sid, remaining);
+    if (ids == NULL) {
+        /* NULL ids means to walk the hash table instead. */
+        starts = 1;
+        /* FIXME: hack to signal hash mode to the crawler thread.
+         * Something more clear would be nice.
+         */
+        crawler_count = -1;
+    } else {
+        /* we allow the autocrawler to restart sub-LRU's before completion */
+        for (int sid = POWER_SMALLEST; sid < POWER_LARGEST; sid++) {
+            if (ids[sid])
+                starts += do_lru_crawler_start(sid, remaining);
+        }
     }
     if (starts) {
+        STATS_LOCK();
+        stats_state.lru_crawler_running = true;
+        stats.lru_crawler_starts++;
+        STATS_UNLOCK();
         pthread_cond_signal(&lru_crawler_cond);
     }
     pthread_mutex_unlock(&lru_crawler_lock);
@@ -646,6 +911,7 @@ enum crawler_result_type lru_crawler_crawl(char *slabs, const enum crawler_run_t
     uint32_t sid = 0;
     int starts = 0;
     uint8_t tocrawl[POWER_LARGEST];
+    bool hash_crawl = false;
 
     /* FIXME: I added this while debugging. Don't think it's needed? */
     memset(tocrawl, 0, sizeof(uint8_t) * POWER_LARGEST);
@@ -653,6 +919,8 @@ enum crawler_result_type lru_crawler_crawl(char *slabs, const enum crawler_run_t
         for (sid = 0; sid < POWER_LARGEST; sid++) {
             tocrawl[sid] = 1;
         }
+    } else if (strcmp(slabs, "hash") == 0) {
+        hash_crawl = true;
     } else {
         for (char *p = strtok_r(slabs, ",", &b);
              p != NULL;
@@ -660,7 +928,6 @@ enum crawler_result_type lru_crawler_crawl(char *slabs, const enum crawler_run_t
 
             if (!safe_strtoul(p, &sid) || sid < POWER_SMALLEST
                     || sid >= MAX_NUMBER_OF_SLAB_CLASSES) {
-                pthread_mutex_unlock(&lru_crawler_lock);
                 return CRAWLER_BADCLASS;
             }
             tocrawl[sid | TEMP_LRU] = 1;
@@ -670,7 +937,7 @@ enum crawler_result_type lru_crawler_crawl(char *slabs, const enum crawler_run_t
         }
     }
 
-    starts = lru_crawler_start(tocrawl, remaining, type, NULL, c, sfd);
+    starts = lru_crawler_start(hash_crawl ? NULL : tocrawl, remaining, type, NULL, c, sfd);
     if (starts == -1) {
         return CRAWLER_RUNNING;
     } else if (starts == -2) {
