@@ -46,6 +46,7 @@
 #endif
 
 #include "itoa_ljust.h"
+#include "slabs_mover.h"
 #include "protocol_binary.h"
 #include "cache.h"
 #include "logger.h"
@@ -358,8 +359,7 @@ struct slab_stats {
 #define PROXY_THREAD_STATS_FIELDS \
     X(proxy_conn_requests) \
     X(proxy_conn_errors) \
-    X(proxy_conn_oom) \
-    X(proxy_req_active)
+    X(proxy_conn_oom)
 #endif
 
 /**
@@ -374,6 +374,7 @@ struct thread_stats {
 #endif
 #ifdef PROXY
     PROXY_THREAD_STATS_FIELDS
+    int64_t proxy_req_active;
 #endif
 #undef X
     struct slab_stats slab_stats[MAX_NUMBER_OF_SLAB_CLASSES];
@@ -394,11 +395,11 @@ struct stats {
     uint64_t      listen_disabled_num;
     uint64_t      slabs_moved;       /* times slabs were moved around */
     uint64_t      slab_reassign_rescues; /* items rescued during slab move */
-    uint64_t      slab_reassign_evictions_nomem; /* valid items lost during slab move */
     uint64_t      slab_reassign_inline_reclaim; /* valid items lost during slab move */
     uint64_t      slab_reassign_chunk_rescues; /* chunked-item chunks recovered */
     uint64_t      slab_reassign_busy_items; /* valid temporarily unmovable */
     uint64_t      slab_reassign_busy_deletes; /* refcounted items killed */
+    uint64_t      slab_reassign_busy_nomem; /* valid items lost during slab move */
     uint64_t      lru_crawler_starts; /* Number of item crawlers kicked off */
     uint64_t      lru_maintainer_juggles; /* number of LRU bg pokes */
     uint64_t      time_in_listen_disabled_us;  /* elapsed time in microseconds while server unable to process new connections */
@@ -432,6 +433,7 @@ struct stats_state {
     uint64_t      curr_bytes;
     uint64_t      curr_conns;
     uint64_t      hash_bytes;       /* size used for hash tables */
+    float         extstore_memory_pressure; /* when extstore might memory evict */
     unsigned int  conn_structs;
     unsigned int  reserved_fds;
     unsigned int  hash_power_level; /* Better hope it's not over 9000 */
@@ -483,7 +485,9 @@ struct settings {
     bool slab_reassign;     /* Whether or not slab reassignment is allowed */
     bool ssl_enabled; /* indicates whether SSL is enabled */
     int slab_automove;     /* Whether or not to automatically move slabs */
+    unsigned int slab_automove_version; /* bump if AM config args change */
     double slab_automove_ratio; /* youngest must be within pct of oldest */
+    double slab_automove_freeratio; /* % of memory to hold free as buffer */
     unsigned int slab_automove_window; /* window mover for algorithm */
     int hashpower_init;     /* Starting hash power level */
     bool shutdown_command; /* allow shutdown command */
@@ -507,6 +511,7 @@ struct settings {
     bool drop_privileges;   /* Whether or not to drop unnecessary process privileges */
     bool watch_enabled; /* allows watch commands to be dropped */
     bool relaxed_privileges;   /* Relax process restrictions when running testapp */
+    struct slab_rebal_thread *slab_rebal; /* struct for page mover thread */
 #ifdef EXTSTORE
     unsigned int ext_io_threadcount; /* number of IO threads to run. */
     unsigned int ext_page_size; /* size in megabytes of storage pages. */
@@ -519,7 +524,6 @@ struct settings {
     unsigned int ext_drop_under; /* when fewer than this many pages, drop COLD items */
     unsigned int ext_max_sleep; /* maximum sleep time for extstore bg threads, in us */
     double ext_max_frag; /* ideal maximum page fragmentation */
-    double slab_automove_freeratio; /* % of memory to hold free as buffer */
     bool ext_drop_unread; /* skip unread items during compaction */
     /* start flushing to extstore after memory below this */
     unsigned int ext_global_pool_min;
@@ -680,29 +684,25 @@ typedef struct {
 #define IO_QUEUE_EXTSTORE 1
 #define IO_QUEUE_PROXY 2
 
+#define IO_PENDING_TYPE_NONE 0
+#define IO_PENDING_TYPE_BASIC 1
+#define IO_PENDING_TYPE_EXTSTORE 2
+#define IO_PENDING_TYPE_PROXY 3
+
 typedef STAILQ_HEAD(iop_head_s, _io_pending_t) iop_head_t;
 typedef struct _io_pending_t io_pending_t;
 typedef struct io_queue_s io_queue_t;
 typedef void (*io_queue_stack_cb)(io_queue_t *q);
 typedef void (*io_queue_cb)(io_pending_t *pending);
-// This structure used to be passed between threads, but is now owned entirely
-// by the worker threads.
 // IO pending objects are created and stacked into this structure. They are
 // then sent off to remote threads.
-// The objects are returned one at a time to the worker threads, and this
-// structure is then consulted to see when to resume the worker.
-struct io_queue_s {
-    void *ctx; // duplicated from io_queue_cb_t
-    void *stack_ctx; // module-specific context to be batch-submitted
-    int count; // ios to process before returning. only accessed by queue processor once submitted
-    int type; // duplicated from io_queue_cb_t
-};
-
-typedef struct io_queue_cb_s {
+// The objects are returned one at a time to the worker threads.
+typedef struct io_queue_s {
     void *ctx; // untouched ptr for specific context
+    iop_head_t stack;
     io_queue_stack_cb submit_cb; // callback given a full stack of pending IO's at once.
     int type;
-} io_queue_cb_t;
+} io_queue_t;
 
 struct thread_notify {
     struct event notify_event;  /* listen event for notify pipe or eventfd */
@@ -724,8 +724,9 @@ typedef struct {
     iop_head_t ion_head;        /* queue for IO object return */
     int cur_sfd;                /* client fd for logging commands */
     int thread_baseid;          /* which "number" thread this is for data offsets */
+    int conns_tosubmit;         /* number of conns which have put data in io_queue */
     struct thread_stats stats;  /* Stats generated by this thread */
-    io_queue_cb_t io_queues[IO_QUEUE_COUNT];
+    io_queue_t io_queues[IO_QUEUE_COUNT];
     struct conn_queue *ev_queue; /* Worker/conn event queue */
     cache_t *rbuf_cache;        /* static-sized read buffers */
     mc_resp_bundle *open_bundle;
@@ -752,8 +753,8 @@ typedef struct {
     int proxy_vm_last_kb;
     unsigned int proxy_vm_negative_delta;
     int proxy_vm_gcrunning;
-    bool proxy_vm_needspoke;
-    uint64_t proxy_active_req_limit;
+    int proxy_vm_gcpokemem;
+    int64_t proxy_active_req_limit;
     uint64_t proxy_buffer_memory_limit; // protected by limit_lock
     uint64_t proxy_buffer_memory_used; // protected by limit_lock
     uint32_t proxy_rng[4]; // fast per-thread rng for lua.
@@ -783,6 +784,7 @@ typedef struct _mc_resp {
      * to asynchronously kill an object that was queued to write
      */
     bool skip;
+    bool suspended; // waiting for response from subsystem
     bool free; // double free detection.
 #ifdef PROXY
     bool proxy_res; // we're handling a proxied response buffer.
@@ -810,7 +812,9 @@ struct _mc_resp_bundle {
 typedef struct conn conn;
 
 struct _io_pending_t {
-    int io_queue_type; // matches one of IO_QUEUE_*
+    uint8_t io_queue_type; // which queue this entered on
+    uint8_t io_sub_type; // sub-type. FIXME: can optimize out later.
+    uint8_t payload;
     LIBEVENT_THREAD *thread;
     conn *c;
     mc_resp *resp; // associated response object
@@ -866,8 +870,7 @@ struct conn {
     /* data for the swallow state */
     int    sbytes;    /* how many bytes to swallow */
 
-    int io_queues_submitted; /* see notes on io_queue_t */
-    io_queue_t io_queues[IO_QUEUE_COUNT]; /* set of deferred IO queues. */
+    int resps_suspended; /* see notes on io_queue_cb_t */
 #ifdef PROXY
     void *proxy_rctx; /* pointer to active request context */
 #endif
@@ -918,27 +921,6 @@ extern volatile bool is_paused;
 extern volatile int64_t delta;
 #endif
 
-/* TODO: Move to slabs.h? */
-extern volatile int slab_rebalance_signal;
-
-struct slab_rebalance {
-    void *slab_start;
-    void *slab_end;
-    void *slab_pos;
-    int s_clsid;
-    int d_clsid;
-    uint32_t busy_items;
-    uint32_t rescues;
-    uint32_t evictions_nomem;
-    uint32_t inline_reclaim;
-    uint32_t chunk_rescues;
-    uint32_t busy_deletes;
-    uint32_t busy_loops;
-    uint8_t done;
-    uint8_t *completed;
-};
-
-extern struct slab_rebalance slab_rebal;
 #ifdef EXTSTORE
 extern void *ext_storage;
 #endif
@@ -954,10 +936,24 @@ enum delta_result_type do_add_delta(LIBEVENT_THREAD *t, const char *key,
                                     item **it_ret);
 enum store_item_type do_store_item(item *item, int comm, LIBEVENT_THREAD *t, const uint32_t hv, int *nbytes, uint64_t *cas, const uint64_t cas_in, bool cas_stale);
 void thread_io_queue_add(LIBEVENT_THREAD *t, int type, void *ctx, io_queue_stack_cb cb);
-void conn_io_queue_setup(conn *c);
-io_queue_t *conn_io_queue_get(conn *c, int type);
-io_queue_cb_t *thread_io_queue_get(LIBEVENT_THREAD *t, int type);
+io_queue_t *thread_io_queue_get(LIBEVENT_THREAD *t, int type);
+void thread_io_queue_submit(LIBEVENT_THREAD *t);
 void conn_io_queue_return(io_pending_t *io);
+#define conn_resp_suspend(c, resp) \
+    do { \
+        resp->suspended = true; \
+        c->resps_suspended++; \
+    } while (0)
+#define conn_resp_unsuspend(c, resp) \
+    do { \
+        resp->suspended = false; \
+        c->resps_suspended--; \
+        assert(c->resps_suspended >= 0); \
+        if (c->resps_suspended == 0) { \
+            conn_worker_readd(c); \
+        } \
+    } while (0)
+
 conn *conn_new(const int sfd, const enum conn_states init_state, const int event_flags, const int read_buffer_size,
     enum network_transport transport, struct event_base *base, void *ssl, uint64_t conntag, enum protocol bproto);
 
